import struct
import wave
import math

class Wave:
  def __init__(self, filename):
    self.w = wave.open(filename, "rb")
    self.samples=self.w.getnframes()
    self.channels=self.w.getnchannels()

  def seek(self, pos):
    self.w.setpos(pos)
    
  def slice(self, start, end=None, buf_size=4096):
    buf=[0]*(buf_size*self.channels)

    if end==None: end=self.samples

    if start>self.samples: return
    if end>self.samples: end=self.samples # clipping
    length=end-start

    self.seek(start)
    while length>buf_size:
      frames=self.w.readframes(buf_size)
      size=len(frames)/(2*self.channels)
      if size==0:
        break
      i=0
      for v in struct.unpack("<%dh" % (size*self.channels), frames):
        if self.channels==1 or (self.channels==2 and i%2==1):
          yield v/16384.0*2
        i+=1
      length-=size

  def __del__(self):
    if self.w: self.w.close()

def zero_crossing(d):
  last_neg=next(d)>0
  last_idx=0
  for idx,sample in enumerate(d):
    neg=sample<0              # negative?
    if neg ^ last_neg:    # any edge (zero crossing)
      yield idx,idx-last_idx
      last_idx=idx
    last_neg=neg         # last was negative

def fsk(d, mark=5327, space=3995, sample_rate=44100):
  # calulations
  fmark=float(sample_rate)/mark/2.0
  fspace=float(sample_rate)/space/2.0
  fdiff=(fspace-fmark)/2.0
  fa,fb,fc=fmark-fdiff,fmark+fdiff,fspace+fdiff
  sz=int(float(sample_rate)/min(mark, space)+0.5)
  buf=20
  #print("fa: %f, fb: %f, fc: %f, sz: %f" % (fa, fb, fc, sz))

  # processing
  b=[0]*buf
  idx=0
  last=-1
  sample_last=-1
  for sample, zero in zero_crossing(d):
    b[idx]=zero
    idx=(idx+1)%buf
    f=float(sum(b))/buf
    v=-1
    if f>=fa and f<=fc:
      v=1 if f>fb else 0

    if v!=last:
      yield sample, v
      sample_last=sample
    last=v

def manchester(d, first=1):
  i=0
  last=first
  for idx, value in zero_crossing(d):
    if value>8:
      last=last^1
      i+=1
    i+=1
    if (i%2)==1: yield idx, last

def irg_manchester(d, min_size=300):
  c=0
  first_idx,old_idx=-1,-1
  zeros=zero_crossing(d)
  for idx, value in zeros:
    if value>8:    # con 9 blue max tien un caso
      if c>1 and idx-first_idx>min_size:
        yield first_idx, idx+1
      c=0
    elif value>3: 
      if c==0: first_idx=idx
      c+=1
    else:
      c=0

    old_idx=idx

def irg_fsk(d, min_size=1000):
  last_idx=-1
  for idx, value in fsk(d):
    if value==1 and idx-last_idx>min_size:
      yield last_idx, idx
    last_idx=idx

def uart_sincronizer(d, bauds=600, sample_rate=44100):
    samples=float(sample_rate)/bauds
 
    v,bit=next(d)
    t=v+(samples/2) # current time: at the middle
    while True:
        if t<(v+samples):
          yield t, bit
          t+=samples
        else:
          t=v+samples+(samples/2) # current time: autosync CTM! (tolerance 562-615 for 600 bauds)
          v,bit=next(d)

def uart_sync(d):
  b,i=0,0
  for idx,v in d:
    if (i==0 and v!=0) or (i==9 and v!=1): return #raise Exception("Bad UART data")
    b=(b>>1)+(v<<8)
    if i==9:
      yield b&0xFF
      b==0
    i=(i+1)%10

def sum8(d):
  s=0
  for v in d:
    s+=v
    s=(s+(s>>8))&0xFF
  return s

def get_fsk_blocks(w, start=0, min_irg=100000):
  blocks=[]
  irgs=list(irg_fsk(w.slice(start), min_irg))
  for start, end in irgs:
    data=list(uart_sync(uart_sincronizer(fsk(w.slice(end)), 600)))
    blocks.append({'irg_start': start, 'irg_end': end, 'data': data, 'checksum': sum8(data[:-1])==data[-1]})
  return blocks

def get_manchester_blocks(w, start=0, min_irg=900):
  blocks=[]
  irgs=list(irg_manchester(w.slice(start), min_irg))
  for start, end in irgs:
    data=list(uart_sync(manchester(w.slice(end), first=0)))
    blocks.append({'irg_start': start, 'irg_end': end, 'data': data, 'checksum': sum8(data[:-1])==data[-1]})
  return blocks

